#!/usr/bin/env python
# This program is public domain

"""
Tree view for NeXus files
-------------------------
The `nexus.tree` routines provide a natural interface to NeXus datasets.
Entries in a group are referenced much like fields in a class are
referenced in python.  Rather than following the directory model of
the `nexus.napi` interface, users are free to reference separate fields
in the dataset at the same time.  Large datasets are not read until 
they are needed, and may be read or written one slab at a time.

There are a number of functions which operate on files::

    import nexus
    nxfile = nexus.load('file.nxs') # loads a structure from a file
    nexus.save('copy.nxs', tree)    # saves a structure to a file
    nexus.tree('copy.nxs')          # display the contents of a file

The tree returned from load() has an entry for each group, field and
attribute.  You can traverse the hierarchy using the names of the
groups.  For example, tree.entry.instrument.detector.distance is an
example of a field containing the distance to each pixel in the 
detector. Entries can also be referenced by NXclass name, such as 
tree.NXentry[0].instrument. Since there may be multiple entries of the 
same NXclass, the NXclass attribute returns a possibly empty list.

Properties of the entry in the tree are referenced by nx attributes.
Depending on the object type, different nx attributes may be available.

Objects (class NXobject) have attributes shared by both groups and fields::
    * name     object name
    * nxclass  object class for groups, 'NXfield' for fields
    * group    group containing the entry, or None for the root
    * attrs    dictionary of NeXus attributes for the object

Groups (class NXgroup) have attributes for accessing children::
    * entries  dictionary of entries within the group
    * component('nxclass')  return group entries of a particular class
    * dir()     print the list of entries in the group
    * tree()    print the list of entries and subentries in the group
    * plot()    plot signal and axes for the group, if available

Fields (class NXfield) have attributes for accessing data:
    * shape     dimensions of data in the field
    * dtype     data type
    * nxdata    data in the field
    * nxdata_as data returned in particular units

Linked fields or groups (class NXlink) have attributes for accessing the link::
    * link      reference to the linked field or group

NeXus attributes (class NXattr) have a type and a value only::
    * dtype     attribute type
    * nxdata    attribute data

Data can be stored in the NeXus file in a variety of units, depending
on which facility is storing the file.  This makes life difficult
for reduction and analysis programs which must know the units they
are working with.  Our solution to this problem is to allow the reader
to retrieve data from the file in particular units.  For example, if
detector distance is stored in the file using millimeters you can 
retrieve them in meters using::

    entry.instrument.detector.distance.nxdata_as('m')

See `nexus.unit` for more details on the unit formats supported.

The slab interface to field data works by opening the file handle
and keeping it open as long as the slab interface is needed.  This
is done in python 2.5 using the with statement.  Once the context
is entered, get() and put() methods on the object allow you to read
and write data a slab at a time.  For example::

    # Read a Ni x Nj x Nk array one vector at a time
    with root.NXentry[0].data.data as slab:
        Ni,Nj,Nk = slab.shape
        size = [1,1,Nk]
        for i in range(Ni):
            for j in range(Nj):
                value = slab.get([i,j,0],size)

The equivalent can be done in Python 2.4 and lower using the context
functions __enter__ and __exit__::

    slab = data.slab.__enter__()
    ... do the slab functions ...
    data.slab.__exit__()

You can traverse the tree by component class instead of component name.  
Since there may be multiple components of the same class in one group
you will need to specify which one to use.  For example, 
tree.NXentry[0].NXinstrument[0].NXdetector[0].distance references the 
first detector of the first instrument of the first entry.  Unfortunately, 
there is no guarantee regarding the order of the entries, and it may vary 
from call to call, so the above is of limited utility.

The plot() method for groups uses matplotlib to plot the data.  You
can replace this with your own plotter by setting nexus.NXgroup._plotter
to your own plotter class.  The plotter class has one method::

    plot(signal, axes, entry, title)

where signal is the field containing the data, axes are the fields
listing the signal sample points, entry is file/path within the file
to the data group and title is the title of the NXentry, if available.

The load() and save() functions are implemented using the class
`nexus.tree.NeXusTree`, a subclass of `nexus.napi.NeXus` which allows 
all the usual API functions.  You can subclass NeXusTree with your
own version that defines, e.g., a NXmonitor() method to return an
NXmonitor object when an NXmonitor class is read.  Your NXmonitor
class should probably be a subclass of NXgroup.
"""

__all__ = ['NeXusTree', 'NXobject', 'NXfield', 'NXgroup', 'NXattr', 
           'NX_MEMORY', 'setmemory', 'load', 'save', 'tree', 'centers', 
           'NXlink', 'NXlinkfield', 'NXlinkgroup', 'SDS', 'NXlinkdata']

from copy import copy

import types
import numpy as np
import napi
import unit
from napi import NeXusError

#Memory in MB
NX_MEMORY = 500

#List of defined base classes (later added to __all__)
_nxclasses = ['NXroot', 'NXentry', 'NXsubentry', 'NXdata', 'NXmonitor',
              'NXlog', 'NXsample', 'NXinstrument', 'NXaperture', 'NXattenuator',
              'NXbeam', 'NXbeam_stop', 'NXbending_magnet', 'NXcharacterization',
              'NXcollection', 'NXcollimator', 'NXcrystal', 'NXdetector',
              'NXdisk_chopper', 'NXenvironment', 'NXevent_data',
              'NXfermi_chopper', 'NXfilter', 'NXflipper', 'NXgeometry',
              'NXguide', 'NXinsertion_device', 'NXmirror', 'NXmoderator',
              'NXmonochromator', 'NXnote', 'NXorientation''NXparameter',
              'NXpolarizer', 'NXpositioner', 'NXprocess', 'NXsensor', 'NXshape',
              'NXsource', 'NXtranslation', 'NXuser', 'NXvelocity_selector']

np.set_printoptions(threshold=5)

class NeXusTree(napi.NeXus):
    """
    Structure-based interface to the NeXus file API.

    Usage::

      file = NeXusTree(filename, ['r','rw','w'])
        - open the NeXus file
      root = file.readfile()
        - read the structure of the NeXus file.  This returns a NeXus tree.
      file.writefile(root)
        - write a NeXus tree to the file.
      data = file.readpath(path)
        - read data from a particular path

    Example::

      nx = NeXusTree('REF_L_1346.nxs','r')
      tree = nx.readfile()
      for entry in tree.NXentry:
          process(entry)
      copy = NeXusTree('modified.nxs','w')
      copy.writefile(tree)

    Note that the large datasets are not loaded immediately.  Instead, the
    when the data set is requested, the file is reopened, the data read, and
    the file closed again.  open/close are available for when we want to
    read/write slabs without the overhead of moving the file cursor each time.
    The NXdata objects in the returned tree hold the object values.

    Subclasses can provide methods for individual NeXus classes such
    as NXbeam or NXdata.  Brave users can also specialize NXgroup, 
    NXattr, NXfield and NXlink methods.
    """
    def readfile(self):
        """
        Read the nexus file structure from the file.  Large datasets
        are not read until they are needed.  

        Returns a tree of NXgroup, NXfield and NXlink objects.
        """
        self.open()
        self.openpath("/")
        root = self._readgroup()
        self.close()
        root.group = None        
        # Resolve links (not necessary now that link is set as a property)
        #self._readlinks(root, root)
        root._file = self
        return root

    def writefile(self, tree):
        """
        Write the nexus file structure to the file.  The file is assumed to
        start empty.

        Updating individual objects can be done using the napi interface, with
        nx.handle as the nexus file handle.
        """
        self.open()
        links = []
        # Root object is special --- only write its children.
        # TODO: maybe want to write root object attributes?
        for entry in tree.entries.values():
            links += self._writegroup(entry, path="")
        self._writelinks(links)
        self.close()

    def readpath(self, path):
        """
        Read the data on a particular file path.

        Returns a numpy array containing the data, a python scalar, or a
        string depending on the shape and storage class.
        """
        self.open()
        self.openpath(path)
        try:
            return self.getdata()
        except ValueError:
            return None

    def _readdata(self, name):
        """
        Read a data object, returning NXfield or NXlink depending on the
        nature of the object.
        """
        # Finally some data, but don't read it if it is big
        # Instead record the location, type and size
        self.opendata(name)
        attrs={}
        attrs = self.getattrs()
        if 'target' in attrs and attrs['target'] != self.path:
            # This is a linked dataset; don't try to load it.
            #print "read link %s->%s"%(path,attrs['target'].nxdata)
            data = NXlinkfield(target=attrs['target'], name=name)
        else:
            dims,type = self.getinfo()
            #Read in the data if it's not too large
            if np.prod(dims) < 1000:# i.e., less than 1k dims
                try:
                    value = self.getdata()
                except ValueError:
                    value = None
            else:
                value = None
            data = NXfield(value=value,name=name,dtype=type,shape=dims,attrs=attrs)
        self.closedata()
        return data

    # These are groups that HDFView explicitly skips
    _skipgroups = ['CDF0.0','_HDF_CHK_TBL_','Attr0.0','RIG0.0','RI0.0',
                   'RIATTR0.0N','RIATTR0.0C']

    def _readchildren(self,n):
        children = {}
        for _item in range(n):
            name,nxclass = self.getnextentry()
            #print "name,class,path",name,nxclass,self.path
            if nxclass in self._skipgroups:
                pass # Skip known bogus classes
            elif nxclass == 'SDS': # NXgetnextentry returns 'SDS' as the class for NXfields
                children[name] = self._readdata(name)
            else:
                self.opengroup(name,nxclass)
                children[name] = self._readgroup()
                self.closegroup()
        return children

    def _readgroup(self):
        """
        Read the currently open group and all subgroups.
        """
        n,name,nxclass = self.getgroupinfo()
        attrs = {}
        attrs = self.getattrs()        
        if 'target' in attrs and attrs['target'] != self.path:
            # This is a linked group; don't try to load it.
            group = self.NXlinkgroup(target=attrs['target'], name=name)
        else:
            children = self._readchildren(n)
            # If we are subclassed with a handler for the particular
            # NXentry class name use that constructor for the group
            # rather than the generic NXgroup class.
#            if hasattr(self,nxclass):
#                factory = getattr(self,nxclass)
#            else:
#                factory = self.NXgroup
            group = NXgroup(nxclass=nxclass,name=name,attrs=attrs,entries=children)
            # Build chain back structure
            for obj in children.values():
                obj.group = group
        return group
    
    def _readlinks(self, root, group):
        """
        Convert linked objects into direct references.
        """
        for entry in group.entries.values():
            if isinstance(entry, NXlink):
                link = root
                try:
                    for level in entry._target[1:].split('/'):
                        link = getattr(link,level)
                    entry.link = link
                except AttributeError:
                    pass
            elif isinstance(entry, NXgroup):
                self._readlinks(root, entry)
                
    def _writeattrs(self, attrs):
        """
        Return the attributes for the currently open group/data or for
        the file if no group or data object is open.
        """
        for name,pair in attrs.iteritems():
            self.putattr(name,pair.nxdata,pair.dtype)

    def _writedata(self, data, path):
        """
        Write the given data.

        NXlinks cannot be written until the linked group is created, so
        this routine returns the set of links that need to be written.
        Call writelinks on the list.
        """

        path = path + "/" + data.name
        #print 'write data',path

        # If the data is linked then
        if hasattr(data,'_target'):
            return [(path, data._target)]

        # Finally some data, but don't read it if it is big
        # Instead record the location, type and size
        #If the array size is too large, their product needs a long integer
        if np.prod(data.shape) > 10000 or np.prod(data.shape) < 0:
            # Compress the fastest moving dimension of large datasets
            slab_dims = np.ones(len(data.shape),'i')
            if data.shape[-1] < 100000:
                slab_dims[-1] = data.shape[-1]
            else: 
                slab_dims[-1] = 100000
            self.compmakedata(data.name, data.dtype, data.shape,
                              'lzw', slab_dims)
        else:
            # Don't use compression for small datasets
            try:
                self.makedata(data.name, data.dtype, data.shape)
            except StandardError,errortype:
                print "Error in tree, makedata: ",errortype
            
        self.opendata(data.name)
        self._writeattrs(data.attrs)
        value = data.nxdata
        if value is not None:
            self.putdata(data.nxdata)
        self.closedata()
        return []

    def _writegroup(self, group, path):
        """
        Write the given group structure, including the data.

        NXlinks cannot be written until the linked group is created, so
        this routine returns the set of links that need to be written.
        Call writelinks on the list.
        """
        path = path + "/" + group.name

        links = []
        self.makegroup(group.name, group.nxclass)
        self.opengroup(group.name, group.nxclass)
        self._writeattrs(group.attrs)
        if hasattr(group, '_target'):
            links += [(path, group._target)]
        for child in group.entries.values():
            if child.nxclass == 'NXfield':
                links += self._writedata(child,path)
            elif hasattr(child,'_target'):
                links += [(path+"/"+child.name,child._target)]
            else:
                links += self._writegroup(child,path)
        self.closegroup()
        return links

    def _writelinks(self, links):
        """
        Create links within the NeXus file as indicated by the set of pairs
        returned by writegroup.
        """
        gid = {}

        # identify targets
        for path,target in links:
            gid[target] = None

        # find gids for targets
        for target in gid.iterkeys():
            #sprint "target",target
            self.openpath(target)
            # Can't tell from the name if we are linking to a group or
            # to a dataset, so cheat and rely on getdataID to signal
            # an error if we are not within a group.
            try:
                gid[target] = self.getdataID()
            except NeXusError:
                gid[target] = self.getgroupID()

        # link sources to targets
        for path,target in links:
            if path != target:
                # ignore self-links
                parent = "/".join(path.split("/")[:-1])
                print "link %s -> %s"%(parent,target)
                self.openpath(parent)
                self.makelink(gid[target])


def _getAxes(axes):
    """
    Parse the 'axes' attribute for the axis names.
    
    The delimiter separating each axis can be white space, a comma, or a colon.
    """
    import re
    sep=re.compile('[\[]*(\s*,*:*)+[\]]*')
    return filter(lambda x: len(x)>0, sep.split(axes))


class NXattr(object):
    """
    Class for NeXus attributes of a NXfield or NXgroup object.
    
    This class is only used for NeXus attributes that are stored in a 
    NeXus file and helps to distinguish them from Python attributes. 
    There are two Python attributes for each NeXus attribute.
    
    Python Attributes
    -----------------
    nxdata : string, Numpy scalar, or Numpy ndarray
        The value of the NeXus attribute.
    dtype : string
        The data type of the NeXus attribute. This is set to 'char' for
        a string attribute or the string of the corresponding Numpy data type
        for a numeric attribute.
    
    NeXus Attributes
    ----------------
    NeXus attributes are stored in the 'attrs' dictionary of the parent object,
    NXfield or NXgroup, but can often be referenced or assigned using the
    attribute name as if it were an object attribute. 
    
    For example, after assigning the NXfield, the following three attribute
    assignments are all equivalent::

        >>> entry.sample.temperature = NXfield(40.0)
        >>> entry.sample.temperature.attrs['units'] = 'K'
        >>> entry.sample.temperature.units = NXattr('K')
        >>> entry.sample.temperature.units = 'K'
    
    The third version above is only allowed for NXfield attributes and is 
    not allowed if the attribute has the same name as one of the following
    internally defined attributes, i.e.,
    
    ['name', 'group', 'entries', 'attrs', 'dtype','shape', 'link'] 
    
    or if the attribute name begins with 'nx' or '_'. It is only possible to 
    reference attributes with one of the proscribed names using the 'attrs'
    dictionary, i.e.,

        >>> entry.sample.group = NXattr('a different group')
        >>> entry.sample.group
        NXgroup('NXentry','entry')
        >>> entry.sample.attrs["group"]
        NXattr('a different group','char')

    """
    def __init__(self,value=None,dtype=''):
        if isinstance(value, NXattr):
            self.nxdata,self.dtype = value.nxdata,value.dtype
        elif dtype:
            if dtype in np.typeDict:
                self.nxdata,self.dtype = np.__dict__[dtype](value),dtype
            elif dtype == 'char':
                self.nxdata,self.dtype = str(value),dtype
            else:
                raise NeXusError, "Invalid data type"
        else:
            if isinstance(value, str):
                self.nxdata,self.dtype = str(value), 'char'
            elif value is not None:
                if isinstance(value, NXobject): 
                    raise NeXusError, "A data attribute cannot be an NXfield or NeXus group"
                else:
                    self.nxdata = np.array(value)
                self.dtype = self.nxdata.dtype.name
                if self.nxdata.size == 1:
                    self.nxdata = np.__dict__[self.dtype](self.nxdata)
            else:
                self.nxdata,self.dtype = None, 'char'

    def __str__(self):
        return str(self.nxdata)

    def __repr__(self):
        if self.dtype == 'char':
            return "NXattr('%s','%s')"%(self.nxdata,self.dtype)
        else:
            return "NXattr(%s,'%s')"%(self.nxdata,self.dtype)

    def __eq__(self, other):
        """
        Return true if the value of the attribute is the same as the other.
        """
        if isinstance(other, NXattr):
            return self.nxdata == other.nxdata
        else:
            return self.nxdata == other

_nxattrs = ['name', 'group', 'entries', 'attrs', 'dtype','shape', 'link', 'path', 'head'] 
_npattrs = filter(lambda x: not x.startswith('_') and x not in  _nxattrs,
                  np.ndarray.__dict__.keys())

class NXobject(object):
    """
    Abstract base class for elements in NeXus files.
    
    The object has a subclass of NXfield, NXgroup, or one of the NXgroup
    subclasses. Child nodes should be accessible directly as object attributes.
    Constructors for NXobject objects are defined by either the NXfield or
    NXgroup classes.
    
    Python Attributes
    -----------------
    nxclass : string
        The class of the NXobject. NXobjects can have class NXfield, NXgroup, or
        be one of the NXgroup subclasses.
    name : string
        The name of the NXobject. Since it is possible to reference the same 
        Python object multiple times, this is not necessarily the same as the
        object name. However, if the object is part of a NeXus tree, this will
        be the attribute name within the tree. 
    group : NXgroup
        The parent group containing this object within a NeXus tree. If the 
        object is not part of any NeXus tree, it will be set to None.
    path : string
        The path to this object with respect to the root of the NeXus tree. For
        NeXus data read from a file, this will be a group of class NXroot, but
        if the NeXus tree was defined interactively, it can be any valid 
        NXgroup.
    head : NXgroup
        The root object of the NeXus tree containing this object. For
        NeXus data read from a file, this will be a group of class NXroot, but
        if the NeXus tree was defined interactively, it can be any valid 
        NXgroup.

    """
    nxclass = "unknown"
    name = "unknown"
    group = None
    _file = None

    def __str__(self):
        return "%s:%s"%(self.nxclass,self.name)

    def __repr__(self):
        return "NXobject('%s','%s')"%(self.nxclass,self.name)
        
    def _setattrs(self, attrs):
        for k,v in attrs.items():
            if isinstance(v,NXattr):
                self.attrs[k] = v
            else:
                self.attrs[k] = NXattr(v)

    def _str_name(self,indent=0):
        if self.nxclass == 'NXfield':
            return " "*indent+self.name
        else:
            return " "*indent+self.name+':'+self.nxclass

    def _str_value(self,indent=0):
        return ""

    def _str_attrs(self,indent=0):
        names = self.attrs.keys()
        names.sort()
        result = []
        for k in names:
            result.append(" "*indent+"@%s = %s"%(k,self.attrs[k].nxdata))
        return "\n".join(result)

    def _str_tree(self,indent=0,attrs=False,recursive=False):
        """
        Print current object and possibly children.
        """
        result = [self._str_name(indent=indent)]
        if attrs and self.attrs: 
            result.append(self._str_attrs(indent=indent+2))
        # Print children
        entries = self.entries
        if entries:
            names = entries.keys()
            names.sort()
            if recursive:
                for k in names:
                    result.append(entries[k]._str_tree(indent=indent+2,
                                                       attrs=attrs, recursive=True))
            else:
                for k in names:
                    result.append(entries[k]._str_name(indent=indent+2))
        result
        return "\n".join(result)
    
    def _getpath(self):
        if self.group is None:
            return ""
        else:
            return self.group._getpath()+"/"+self.name
    
    def _getheadgroup(self):
        if self.group is None:
            return self
        else:
            return self.group._getheadgroup()
    
    path = property(_getpath, "Path to NeXus object")
    head = property(_getheadgroup, "Head group of this object's tree")

    def dir(self,attrs=False,recursive=False):
        """
        Print the object directory.
        
        The directory is a list of NeXus objects within this object, either 
        NeXus groups or NXfields. If 'attrs' is True, NXfield attributes are 
        displayed. If 'recursive' is True, the contents of child groups are 
        also displayed.
        """
        print self._str_tree(attrs=attrs,recursive=recursive)

    def tree(self,attrs=True):
        """
        Print the directory tree.
        
        The tree contains all child objects of this object and their children. 
        It invokes the 'dir' method with both 'attrs' and 'recursive' set 
        to True.
        """
        print self._str_tree(attrs=attrs,recursive=True)
        
    def treestr(self, attrs=True):
        """Return directory tree string"""
        return self._str_tree(attrs=attrs,recursive=True)
    
    def save(self, filename=None, format='w5'):
        """
        Save the NeXus object to a data file.
        
        The object is wrapped in an NXroot group (with name 'root') and an 
        NXentry group (with name 'entry'), if necessary, in order to produce 
        a valid NeXus file.
        """
        if self.nxclass == "NXroot":
            tree = self
        elif self.nxclass == "NXentry":
            tree = NXroot(self)
        else:
            tree = NXroot(NXentry(self))
        file = NeXusTree(filename, format)
        file.writefile(tree)
        file.close()
        tree._file = file
        tree._file.open()
        tree._file.openpath('/')
        tree._setattrs(tree._file.getattrs())
        tree._file.close()


class NXfield(NXobject):
    """
    A NeXus data field.
    
    This is a subclass of NXobject that contains scalar, array, or string data
    and associated NeXus attributes. 
    
    NXfield(value=None, name='unknown', dtype='', shape=[], attrs={}, file=None, 
            path=None, group=None, **attr)

    Input Parameters
    ----------------
    value : scalar value, Numpy array, or string
        The numerical or string value of the NXfield, which is directly
        accessible as the NXfield attribute 'nxdata'.
    name : string
        The name of the NXfield, which is directly accessible as the NXfield
        attribute 'name'. If the NXfield is initialized as the attribute of a
        parent object, the name is automatically set to the name of this
        attribute.
    dtype : string
        The data type of the NXfield value, which is directly accessible as the
        NXfield attribute 'dtype'. Valid input types correspond to standard
        Numpy data types, using names defined by the NeXus API,
        i.e., 'float32' 'float64'
              'int8' 'int16' 'int32' 'int64'
              'uint8' 'uint16' 'uint32' 'uint64'
              'char'
        If the data type is not specified, then it is determined automatically
        by the data type of the 'value' parameter.
    shape : list of ints
        The dimensions of the NXfield data, which is accessible as the NXfield
        attribute 'shape'. This corresponds to the shape of the Numpy array. 
        Scalars (numeric or string) are stored as Numpy zero-rank arrays,
        for which shape=[].
    attrs : dict
        A dictionary containing NXfield attributes. The dictionary values should
        all have class NXattr.
    file : filename
        The file from which the NXfield has been read.
    path : string
        The path to this object with respect to the root of the NeXus tree, 
        using the convention for unix file paths.
    group : NXgroup or subclass of NXgroup
        The parent NeXus object. If the NXfield is initialized as the attribute
        of a parent group, this attribute is automatically set to the parent group.

    Python Attributes
    -----------------
    nxclass : 'NXfield'
        The class of the NXobject.
    name : string
        The name of the NXfield. Since it is possible to reference the same 
        Python object multiple times, this is not necessarily the same as the
        object name. However, if the field is part of a NeXus tree, this will
        be the attribute name within the tree. 
    group : NXgroup
        The parent group containing this field within a NeXus tree. If the 
        field is not part of any NeXus tree, it will be set to None.
    dtype : string or Numpy dtype
        The data type of the NXfield value. If the NXfield has been initialized
        but the data values have not been read in or defined, this is a string.
        Otherwise, it is set to the equivalent Numpy dtype. 
    shape : list or tuple of ints
        The dimensions of the NXfield data. If the NXfield has been initialized
        but the data values have not been read in or defined, this is a list of
        ints. Otherwise, it is set to the equivalent Numpy shape, which is a
        tuple. Scalars (numeric or string) are stored as Numpy zero-rank arrays,
        for which shape=().
    attrs : dictionary
        A dictionary of all the NeXus attributes associated with the field. 
        These are objects with class NXattr.
    nxdata : scalar, Numpy array or string
        The data value of the NXfield. This is normally initialized using the 
        'value' parameter (see above). If the NeXus data is contained 
        in a file and the size of the NXfield array is too large to be stored 
        in memory, the value is not read in until this attribute is directly
        accessed. Even then, if there is insufficient memory, a value of None
        will be returned. In this case, the NXfield array should be read as a 
        series of smaller slabs using 'get'. 
    nxdata_as('units') : scalar value or Numpy array
        If the NXfield 'units' attribute has been set, the data values, stored
        in 'nxdata', are returned after conversion to the specified units.
    path : string
        The path to this object with respect to the root of the NeXus tree. For
        NeXus data read from a file, this will be a group of class NXroot, but
        if the NeXus tree was defined interactively, it can be any valid 
        NXgroup.
    head : NXgroup
        The root object of the NeXus tree containing this object. For
        NeXus data read from a file, this will be a group of class NXroot, but
        if the NeXus tree was defined interactively, it can be any valid 
        NXgroup.

    NeXus Attributes
    ----------------
    NeXus attributes are stored in the 'attrs' dictionary of the NXfield, but
    can usually be assigned or referenced as if they are Python attributes, as
    long as the attribute name is not the same as one of those listed above.
    This is to simplify typing in an interactive session and should not cause
    any problems because there is no name clash with attributes so far defined
    within the NeXus standard. When writing modules, it is recommended that the
    attributes always be referenced using the 'attrs' dictionary if there is
    any doubt.
    
    a) Assigning a NeXus attribute
    
    In the example below, after assigning the NXfield, the following three
    NeXus attribute assignments are all equivalent:

        >>> entry.sample.temperature = NXfield(40.0)
        >>> entry.sample.temperature.attrs['units'] = 'K'
        >>> entry.sample.temperature.units = NXattr('K')
        >>> entry.sample.temperature.units = 'K'
    
    The third version should be avoided if the NeXus attribute has the same name
    as one of the Python attributes defined above or if the attribute name
    begins with '_'.
    
    b) Referencing a NeXus attribute
    
    If the name of the NeXus attribute is not the same as any of the Python
    attributes listed above or any of the attributes defined for Numpy arrays,
    they can be referenced as if they were a Python attribute of the NXfield.   
    It is only possible to reference attributes with one of the proscribed names
    using the 'attrs' dictionary, i.e.,

        >>> entry.sample.temperature.group = NXattr('a different group')
        >>> entry.sample.temperature.group
        NXgroup('NXsample','sample')
        >>> entry.sample.temperature.attrs["group"]
        NXattr('a different group','char')

    Numerical Operations on NXfields
    --------------------------------
    NXfields usually consist of arrays of numeric data with associated
    meta-data, the NeXus attributes. The exception is when they contain
    character strings. This makes them similar to Numpy arrays, and this module
    allows the use of NXfields in numerical operations in the same way as Numpy
    ndarray objects. NXfields are technically not a sub-class of the ndarray
    class, but most Numpy operations work on NXfields, returning either another
    NXfield or, in some cases, an ndarray that can easily be converted to an
    NXfield.
    
        >>> x = NXfield((1.0,2.0,3.0,4.0),name='x')
        >>> print x+1
        [ 2.  3.  4.  5.]
        >>> print 2*x
        [ 2.  4.  6.  8.]
        >>> print x/2
        [ 0.5  1.   1.5  2. ]
        >>> print x**2
        [  1.   4.   9.  16.]
        >>> print x.reshape((2,2))
        [[ 1.  2.]
         [ 3.  4.]]
        >>> y = NXfield((0.5,1.5,2.5,3.5))
        >>> x+y
        NXfield(name=x,value=[ 1.5  3.5  5.5  7.5])
        >>> x*y
        NXfield(name=x,value=[  0.5   3.    7.5  14. ])
        >>> (x+y).shape
        (4,)
        >>> (x+y).dtype
        dtype('float64')

    All these operations return valid NXfield objects containing the same
    attributes as the first NXobject in the expression. The 'reshape' and 
    'transpose' methods also return NXfield objects.
    
    It is possible to use the standard slice syntax.

        >>> x=NXfield(np.linspace(0,10,11),name='x')
        >>> x
        NXfield(name=x,value=[  0.   1.   2. ...,   8.   9.  10.])
        >>> x[2:5]
        NXfield(name=x,value=[ 2.  3.  4.])
    
    In addition, it is possible to use floating point numbers as the slice
    indices. If one of the indices is not integer, both indices are used to
    extract elements in the array with values between the two index values.
    
        >>> x=NXfield(np.linspace(0,100.,11),name='x')
        >>> x
        NXfield(name=x,value=[   0.   10.   20. ...,   80.   90.  100.])
        >>> x[20.:50.]
        NXfield(name=x,value=[ 20.  30.  40.  50.])
   
    The standard Numpy ndarray attributes and methods will also work with 
    NXfields, but will return scalars or Numpy arrays.
    
        >>> x.size
        4
        >>> x.sum()
        10.0
        >>> x.max()
        4.0
        >>> x.mean()
        2.5
        >>> x.var()
        1.25
        >>> x.reshape((2,2)).sum(1)
        array([ 3.,  7.])
    
    Finally, NXfields are cast as ndarrays for operations that require them.
    The returned value will be the same as for the equivalent ndarray 
    operation, e.g.,
    
    >>> np.sin(x)
    array([ 0.84147098,  0.90929743,  0.14112001, -0.7568025 ])
    >>> np.sqrt(x)
    array([ 1.        ,  1.41421356,  1.73205081,  2.        ])
    
    Methods
    -------
    dir(self, attrs=False):
        Print the NXfield specification.
        
        This outputs the name, dimensions and data type of the NXfield. 
        If 'attrs' is True, NXfield attributes are displayed.

    tree(self, attrs=True):
        Print the NXfield and its attributes.
        
        It invokes the 'dir' method with 'attrs' set to True.
        
    save(self, filename, format='w5')
        Save the NXfield into a file wrapped in a NXroot group and NXentry group
        with default names. This is equivalent to 
        
        >>> NXroot(NXentry(NXfield(...))).save(filename)

    Examples
    --------
    >>> x = NXfield(np.linspace(0,2*np.pi,101), units='degree')
    >>> phi = x.nxdata_as(units='radian')
    >>> y = NXfield(np.sin(phi))
    
    # Read a Ni x Nj x Nk array one vector at a time
    >>> with root.NXentry[0].data.data as slab:
            Ni,Nj,Nk = slab.shape
            size = [1,1,Nk]
            for i in range(Ni):
                for j in range(Nj):
                    value = slab.get([i,j,0],size)

    """
    def __init__(self, value=None, name='unknown', dtype='', shape=[], attrs={}, group=None, 
                 **attr):
        if isinstance(value, list) or isinstance(value, tuple):
            value = np.array(value)
        self._value = value
        self.nxclass = 'NXfield'
        self.name = name
        self.dtype = dtype
        self.shape = list(shape)
        self.group = group
        # Append extra keywords to the attribute list
        self.attrs = {}
        for key in attr.keys():
            attrs[key] = attr[key]
        # Convert NeXus attributes to python attributes
        self._setattrs(attrs)
        if 'units' in attrs:
            units = attrs['units']
        else:
            units = None
        self._converter = unit.Converter(units)
        self._incontext = False
        del attrs
        if value is not None and dtype == 'char': value = str(value)
        self._nxdata_set(value)

    def __repr__(self):
        if self._value is not None:
            return "NXfield(name=%s,value=%s)" % (self.name, self._str_value())
        else:
            return "NXfield(name='%s')" % self.name
        
    def __getattr__(self, name):
        """
        Enable standard numpy ndarray attributes if not otherwise defined.
        """
        if name in _npattrs:
            return self.nxdata.__getattribute__(name)
        elif name in self.attrs:
            return self.attrs[name].nxdata
        raise KeyError(name+" not in "+self.name)

    def __setattr__(self, name, value):
        """
        Add an attribute to the NXfield 'attrs' dictionary unless the attribute
        name starts with 'nx' or '_', or unless it is one of the standard Python
        attributes for the NXfield class.
        """
        if isinstance(value, NXattr):
            self.attrs[name] = value
        elif name.startswith('nx') or name.startswith('_') or name in _nxattrs:
            object.__setattr__(self, name, value)
        else:
            self.attrs[name] = NXattr(value)

    def __len__(self):
        """
        Return the length of the NXfield data.
        """
        if self._value is None:
            return np.prod(self.shape)
        else:
            if self.nxdata.dtype.name.startswith('string'):
                return self.nbytes
            else:
                return self.nxdata.size
        
    def index(self, value, max=False):
        """
        Return the index of the NXfield nxdata array that is greater than or equal to the value.
        
        If max, then return the index that is less than or equal to the value.
        This should only be used on one-dimensional monotonically increasing arrays.
        """
        if max:
            return len(self.nxdata)-len(self.nxdata[self.nxdata>=value])
        else:
            return len(self.nxdata[self.nxdata<value])

    def __getitem__(self, index):
        """
        Returns a slice from the NXfield.
        
        In most cases, the slice values are applied to the NXfield nxdata array
        and returned within an NXfield object with the same metadata. However,
        if the array is one-dimensional and the index start and stop values 
        are real, the nxdata array is returned with values between those limits.
        This is to allow axis arrays to be limited by their actual value. This
        real-space slicing should only be used on monotonically increasing (or
        decreasing) one-dimensional arrays.
        """
        if isinstance(index, slice) and \
           (isinstance(index.start, float) or isinstance(index.stop, float)):
            index = slice(self.index(index.start), self.index(index.stop,max=True)+1)
        if self._value is not None:
            result = self.nxdata.__getitem__(index)
        else:
            offset = np.zeros(len(self.shape),dtype=int)
            size = np.array(self.shape)
            if isinstance(index, int):
                offset[0] = index
                size[0] = 1
            else:
                if isinstance(index, slice): index = [index]
                i = 0
                for ind in index:
                    if isinstance(ind, int):
                        offset[i] = ind
                        size[i] = 1
                    else:
                        if ind.start: offset[i] = ind.start
                        if ind.stop: size[i] = ind.stop - offset[i]
                    i = i + 1
            try:
                result = self.get(offset, size)
            except ValueError:
                result = self.nxdata.__getitem__(index)
        return NXfield(result, name=self.name, attrs=self.attrs)
                
    def __setitem__(self, index, value):
        """
        Assign a slice to the NXfield.
        """
        if self._value is not None:
            self.nxdata[index] = value
        else:
            raise NeXusError, "NXfield dataspace not yet allocated"
                
    def __array__(self):
        """
        Cast the NXfield as an array when it is expected by numpy
        """
        return self.nxdata
                
    def __eq__(self, other):
        """
        Return true if the values of the NXfield are the same.
        """
        if isinstance(other, NXfield):
            if isinstance(self.nxdata, np.ndarray) and isinstance(other.nxdata, np.ndarray):
                return all(self.nxdata == other.nxdata)
            else:
                return self.nxdata == other.nxdata
        else:
            return False

    def __ne__(self, other):
        """
        Return true if the values of the NXfield are not the same.
        """
        if isinstance(other, NXfield):
            if isinstance(self.nxdata, np.ndarray) and isinstance(other.nxdata, np.ndarray):
                return any(self.nxdata != other.nxdata)
            else:
                return self.nxdata != other.nxdata
        else:
            return True

    def __add__(self, other):
        """
        Return the sum of the NXfield and another NXfield or number.
        """
        try:
            if isinstance(other, NXfield):
                return NXfield(value=self.nxdata+other.nxdata, name=self.name, 
                               attrs=self.attrs)
            else:
                return NXfield(value=self.nxdata+other, name=self.name, 
                               attrs=self.attrs)
        except TypeError, message:
            raise NeXusError, message

    def __radd__(self, other):
        """
        Return the sum of the NXfield and another NXfield or number.
        
        This variant makes __add__ commutative.
        """
        return self.__add__(other)

    def __sub__(self, other):
        """
        Return the NXfield with the subtraction of another NXfield or number.
        """
        try:
            if isinstance(other, NXfield):
                return NXfield(value=self.nxdata-other.nxdata, name=self.name, 
                               attrs=self.attrs)
            else:
                return NXfield(value=self.nxdata-other, name=self.name, 
                               attrs=self.attrs)
        except TypeError, message:
            raise NeXusError, message

    def __mul__(self, other):
        """
        Return the product of the NXfield and another NXfield or number.
        """
        try:
            if isinstance(other, NXfield):
                return NXfield(value=self.nxdata*other.nxdata, name=self.name, 
                               attrs=self.attrs)
            else:
                return NXfield(value=self.nxdata*other, name=self.name, 
                               attrs=self.attrs)
        except TypeError, message:
            raise NeXusError, message

    def __rmul__(self, other):
        """
        Return the product of the NXfield and another NXfield or number.
        
        This variant makes __mul__ commutative.
        """
        return self.__mul__(other)

    def __div__(self, other):
        """
        Return the NXfield divided by another NXfield or number.
        """
        try:
            if isinstance(other, NXfield):
                return NXfield(value=self.nxdata/other.nxdata, name=self.name, 
                               attrs=self.attrs)
            else:
                return NXfield(value=self.nxdata/other, name=self.name, 
                               attrs=self.attrs)
        except TypeError, message:
            raise NeXusError, message

    def __rdiv__(self, other):
        """
        Return the inverse of the NXfield divided by another NXfield or number.
        """
        try:
            if isinstance(other, NXfield):
                return NXfield(value=other.nxdata/self.nxdata, name=self.name, 
                               attrs=self.attrs)
            else:
                return NXfield(value=other/self.nxdata, name=self.name, 
                               attrs=self.attrs)
        except TypeError, message:
            raise NeXusError, message

    def __pow__(self, power):
        """
        Return the NXfield raised to the specified power.
        """
        try:
            return NXfield(value=pow(self.nxdata,power), name=self.name, 
                        attrs=self.attrs)
        except TypeError, message:
            raise NeXusError, message

    def reshape(self, shape):
        """
        Returns an NXfield with shape.
        """
        self.shape = list(shape)
        if isinstance(self.nxdata, np.ndarray): self.nxdata.shape = shape
        return self

    def transpose(self,shape=None):#added shape variable here 9/8/2010
        """
        Returns an NXfield containing the transpose of the data array. Equivalent 
        to the Numpy ndarray.transpose.
        """
        # error, no shape variable in this scope
        return NXfield(value=self.nxdata.transpose(shape), name=self.name, 
                       attrs=self.attrs)

    def centers(self):
        """
        Returns an NXfield with the centers of a single axis 
        assuming it contains bin boundaries.
        """
        return NXfield((self.nxdata[:-1]+self.nxdata[1:])/2,
                        name=self.name,attrs=self.attrs)

    def __enter__(self):
        """
        Open the datapath for reading slab by slab.

        Note: the results are undefined if you try accessing
        more than one slab at a time.  Don't nest your 
        "with data" statements!
        """
        # TODO: provide a file lock to prevent movement of the
        # file cursor when in the slab context.
        # TODO: if HDF allows multiple cursors, extend napi to support them
        self._close_on_exit = not self.head._file.isopen
        self.head._file.open() # Force file open even if closed
        self.head._file.openpath(self.path)
        self._incontext = True
        return self

    def __exit__(self, *args):
        """
        Close the file associated with the data after reading.
        """
        self._incontext = False
        if self._close_on_exit:
            self.head._file.close()

    def get(self, offset, size, units=""):
        """
        Return a slab from the data array.

        Offsets are 0-origin. Shape can be inferred from the data.
        Offset and shape must each have one entry per dimension.
        
        If units are specified, convert the values to the given units
        before returning them.

        This operation should be performed in a "with group.data"
        context.

        Raises ValueError cannot convert units.

        Corresponds to NXgetslab(handle,data,offset,shape)
        """
        if self.head._file:
            self.__enter__()
            value = self.head._file.getslab(offset,size)
            self.__exit__()
            return self._converter(value,units)
        else:
            raise NeXusError, "No input file specified for NXgetslab"

    def put(self, data, offset, refresh=True):
        """
        Put a slab into the data array.

        Offsets are 0-origin.  Shape can be inferred from the data.
        Offset and shape must each have one entry per dimension.

        This operation should be performed in a "with group.data"
        context.

        Raises NeXusError if this fails.  No error is raised when
        writing to a file which is open read-only.

        Corresponds to NXputslab(handle,data,offset,shape)
        """
        if self.head._file:
            self.__enter__()
            if isinstance(data, NXfield):
                self.head._file.putslab(data.nxdata.astype(self.dtype), offset, data.shape)
            else:
                self.head._file.putslab(data.astype(self.dtype), offset, data.shape)
            self.__exit__()
            if refresh: self.refresh()
        else:
            raise NeXusError, "No output file specified for NXputslab"
        
    def add(self, data, offset):
        """
        Add a slab into the data array.

        Calls get to read in existing data before adding the value
        and calling put. It assumes that the two sets of data have
        compatible data types.
        """
        if isinstance(data, NXfield):
            value = self.get(offset, data.shape)        
            self.put(data.nxdata.astype(self.dtype)+value, offset)
        else:
            value = self.get(offset, data.shape)
            self.put(data.astype(self.dtype)+value, offset)
        
    def refresh(self):
        """
        Rereads the data from the file.

        Calls net to read in data again. If put has been called, then
        nxdata is no longer synchronized with the file making a refresh
        necessary. This will only be performed if nxdata already stores the 
        data.
        """
        if self._value is not None:
            if self.head._file:
                self._value = self.head._file.readpath(self.path)
            else:
                raise NeXusError, "No file specified for reads"
        
    def __str__(self):
        """
        If value is loaded, return the value as a string.  If value is
        not loaded, return the empty string.  Only the first view values
        for large arrays will be printed.
        """
        if self._value is not None:
            return str(self._value)
        return ""

    def _str_value(self,indent=0):
        v = str(self)
        if '\n' in v:
            v = '\n'.join([(" "*indent)+s for s in v.split('\n')])
        return v

    def _str_tree(self,indent=0,attrs=False,recursive=False):
        dims = 'x'.join([str(n) for n in self.shape])
        s = str(self)
        if '\n' in s or s == "":
            s = "%s(%s)"%(self.dtype, dims)
        v=[" "*indent + "%s = %s"%(self.name, s)]
        if attrs and self.attrs: v.append(self._str_attrs(indent=indent+2))
        return "\n".join(v)

    def nxaxes(self):
        """
        Return a list of NXfields containing axes.
        
        Only works if the NXfield has the 'axes' attribute
        """
        try:
            return [getattr(self.group,name) for name in _getAxes(self.axes)]
        except KeyError:
            return None

    def nxdata_as(self, units=""):
        """
        Return the data in particular units.
        """
        if self._value is None:
            if self.head._file:
                if np.prod(self.shape) * np.dtype(self.dtype).itemsize <= NX_MEMORY*1024*1024:
                    self._value = self.head._file.readpath(self.path)
                else:
                    raise MemoryError, 'Data size larger than NX_MEMORY=%s MB' % NX_MEMORY
            else:
                return None
                
        return self._converter(self._value,units)
    
    def _nxdata_set(self, value):
        if value is not None:
            if self.dtype in np.typeDict:
                self._value = np.array(value,self.dtype)
            else:
                self._value = np.array(value)
            self.shape = self._value.shape
            self.dtype = self._value.dtype

    nxdata = property(nxdata_as,_nxdata_set,doc="The data values in default units")
    size = property(__len__,doc="The total size of the data")

SDS = NXfield # For backward compatibility

class PylabPlotter(object):
    """
    Matplotlib plotter object for NeXus data.
    """
    def plot(self, signal, axes, title, errors=None, **opts):
        """
        Plot the data entry.
        
        Raises NeXusError if the data cannot be plotted.
        """
        import pylab
        if "over" in opts.keys():
            over = True
            del opts["over"]
        else:
            over = False
        if not over: pylab.clf()
        
        if "log" in opts.keys():
            logplot = True
            del opts["log"]
        else:
            logplot = False

        # Find the centers of the bins for histogrammed data
        axis_data = centers(signal, axes)

        #One-dimensional Plot
        if len(signal.shape) == 1:
            if hasattr(signal, 'units'):
                if not errors and signal.units == 'counts':
                    errors = NXfield(np.sqrt(signal))
            if logplot:
                data = np.log10(np.clip(signal,0,1e8))
                if errors: ebars = np.log10(errors)
            else:
                data = signal
                if errors: ebars = errors
            if errors:
                myopts=copy(opts)
                myopts.setdefault('fmt','o')
                myopts.setdefault('linestyle','None')
                pylab.scatter(axis_data[0], signal, **opts)
                pylab.errorbar(axis_data[0], signal, ebars, **myopts)
            else:
                pylab.scatter(axis_data[0], signal, **opts)
            if not over:
                pylab.xlabel(label(axes[0]))
                pylab.ylabel(label(signal))
                pylab.title(title)
 
        #Two dimensional plot
        else:
            #from api.nexus import meshgl
            #gridplot = meshgl.pcolor_gl
            #gridplot = pylab.pcolormesh
            if len(signal.shape) == 2:
                data = signal
            else:
                slab = [slice(None), slice(None)]
                for _dim in signal.shape[2:]:
                    slab.append(0)
                data = signal[slab].view().reshape(signal.shape[:2])
                print "Warning: Only the top 2D slice of the data is plotted"
            gridplot = imshow_irregular
#            if hasattr(signal, 'units') and signal.units == 'counts':
            if logplot:
                gridplot(axis_data[0], axis_data[1], 
                         np.log10(np.clip(data,0.,1e8)+1).T, **opts)
            else:
                gridplot(axis_data[0], axis_data[1], np.clip(data,-1e8,1e8).T, **opts)
            pylab.xlabel(label(axes[0]))
            pylab.ylabel(label(axes[1]))
            pylab.title(title)

        #For higher-dimensional plots, choose limits
        #else:
            #raise NeXusError, "Only one and two-dimensional plots currently supported"

    @staticmethod
    def show():
        import pylab
        pylab.show()    

class NXgroup(NXobject):
    """
    A NeXus group object.
    
    This is a subclass of NXobject and is the base class for the specific
    NeXus group classes, e.g., NXentry, NXsample, NXdata. 
    
    NXgroup(*items, **opts)

    Parameters
    ----------
    The NXgroup parameters consist of a list of positional and/or keyword 
    arguments.
    
    Positional Arguments : These must be valid NeXus objects, either an NXfield 
    or a NeXus group. These are added without modification as children of this 
    group. 
    
    Keyword Arguments : Apart from a list of special keywords shown below, 
    keyword arguments are used to add children to the group using the keywords
    as attribute names. The values can either be valid NXfields or NXgroups, 
    in which case the 'name' attribute is changed to the keyword, or they 
    can be numerical or string data, which are converted to NXfield objects.
    
    Special Keyword Arguments:

    name : string
        The name of the NXgroup, which is directly accessible as the NXgroup 
        attribute 'name'. If the NXgroup is initialized as the attribute of 
        a parent group, the name is automatically set to the name of this 
        attribute. If 'nxclass' is specified and has the usual prefix 'NX',
        the default name is the class name without this prefix.
    nxclass : string
        The class of the NXgroup. 
    entries : dict
        A dictionary containing a list of group entries. This is an 
        alternative way of adding group entries to the use of keyword 
        arguments.
    file : filename
        The file from which the NXfield has been read.
    path : string
        The path to this object with respect to the root of the NeXus tree, 
        using the convention for unix file paths.
    group : NXobject (NXgroup or subclass of NXgroup)
        The parent NeXus group, which is accessible as the group attribute 
        'group'. If the group is initialized as the attribute of 
        a parent group, this is set to the parent group.

    Python Attributes
    -----------------
    nxclass : string
        The class of the NXobject.
    name : string
        The name of the NXfield. 
    entries : dictionary
        A dictionary of all the NeXus objects contained within an NXgroup.
    attrs : dictionary
        A dictionary of all the NeXus attributes, i.e., attribute with class NXattr.
    entries : dictionary
        A dictionary of all the NeXus objects contained within the group.
    attrs : dictionary
        A dictionary of all the group's NeXus attributes, which all have the
        class NXattr.
    path : string
        The path to this object with respect to the root of the NeXus tree. For
        NeXus data read from a file, this will be a group of class NXroot, but
        if the NeXus tree was defined interactively, it can be any valid 
        NXgroup.
    head : NXgroup
        The root object of the NeXus tree containing this object. For
        NeXus data read from a file, this will be a group of class NXroot, but
        if the NeXus tree was defined interactively, it can be any valid 
        NXgroup.

    NeXus Group Entries
    -------------------
    Just as in a NeXus file, NeXus groups can contain either data or other
    groups, represented by NXfield and NXgroup objects respectively. To
    distinguish them from regular Python attributes, all NeXus objects are
    stored in the 'entries' dictionary of the NXgroup. However, they can usually
    be assigned or referenced as if they are Python attributes, i.e., using the
    dictionary name directly as the group attribute name, as long as this name
    is not the same as one of the Python attributes defined above or as one of
    the NXfield Python attributes.  
    
    a) Assigning a NeXus object to a NeXus group
    
    In the example below, after assigning the NXgroup, the following three
    NeXus object assignments are all equivalent:

        >>> entry.sample = NXsample()
        >>> entry.sample.entries['temperature'] = NXfield(40.0,name='temperature')
        >>> entry.sample.temperature = NXfield(40.0)
        >>> entry.sample.temperature = 40.0
        >>> entry.sample.temperature
        NXfield(name=temperature,value=40.0)
        
    If the assigned value is not a valid NXobject, then it is cast as an NXfield 
    with the name as the attribute name and a type determined from the Python
    data type.
    
        >>> entry.sample.temperature = 40.0
        >>> entry.sample.temperature
        NXfield(name=temperature,value=40.0)
        >>> entry.data.data.x=np.linspace(0,10,11).astype('float32')
        >>> entry.data.data.x
        NXfield(name=x,value=[  0.   1.   2. ...,   8.   9.  10.])
    
    If the NeXus object has the same name as one of the Python attributes
    defined above, only the first version is allowed. Note, however, that it is
    essential to specify the value of the 'name' keyword to match the dictionary
    name. Since this is inconvenient, it is recommended that the insert module 
    is used in these cases, e.g., to add a NeXus field with the name 'group',
    
        >>> entry.sample.insert(NXfield(40.0, name='group'))
        >>> entry.sample.group
        NXgroup('NXentry','entry')
        >>> entry.sample.entries['group']
        NXfield(name=group,value=40.0)
    
    b) Referencing a NeXus object in a NeXus group
    
    If the name of the NeXus object is not the same as any of the Python
    attributes listed above, they can be referenced as if they were a Python
    attribute of the NXgroup. It is only possible to reference attributes with
    one of the proscribed names using the 'entries' dictionary, i.e.,

        >>> entry.sample.temperature = NXfield(40.0)
        NXfield(name=temperature,value=40.0)
    
    Numerical operations can then be applied directly, as described in the 
    NXfield documentation.
    
        >>> entry.sample.temperature + 10.0
        NXfield(name=temperature,value=50.0)

    NeXus Attributes
    ----------------
    NeXus attributes are not currently used much with NXgroups, except for the
    root group, which has a number of global attributes to store the file name,
    file creation time, and NeXus and HDF version numbers. However, the
    mechanism described for NXfields works here as well. All NeXus attributes
    are stored in the 'attrs' dictionary of the NXgroup, but can be referenced
    as if they are Python attributes as long as there is no name clash.

    Methods
    -------
    insert(self, NXobject, name='unknown'):
        Insert a valid NXobject (NXfield or NXgroup) into the group.
        
        If NXobject has a 'name' attribute and the 'name' keyword is not given,
        then the object is inserted with the NXobject name.
        
    makelink(self, NXobject):
        Add the NXobject to the group entries as a link (NXlink).

    dir(self, attrs=False):
        Print the group directory.
        
        The directory is a list of NeXus objects within this group, either NeXus
        groups or NXfield data. If 'attrs' is True, NXfield attributes are
        displayed. If 'recursive' is True, the contents of child groups are also
        displayed.

    tree(self, attrs=True):
        Print the NXfield and its attributes.
        
        It invokes the 'dir' method with both 'attrs' and 'recursive' 
        set to True.
        
    save(self, filename, format='w5')
        Save the NeXus group into a file

        The object is wrapped in an NXroot group (with name 'root') and an 
        NXentry group (with name 'entry'), if necessary, in order to produce 
        a valid NeXus file.

    Examples
    --------
    >>> x = NXfield(np.linspace(0,2*np.pi,101), units='degree')
    >>> entry = NXgroup(x, name='entry', nxclass='NXentry')
    >>> entry.sample = NXgroup(temperature=NXfield(40.0,units='K'),
                               nxclass='NXsample')
    >>> entry.sample.tree()
    sample:NXsample
      temperature = 40.0
        @units = K

    Note: All the currently defined NeXus classes are defined as subclasses
          of the NXgroup class. It is recommended that these are used 
          directly, so that the above examples become:

    >>> entry = NXentry(x)
    >>> entry.sample = NXsample(temperature=NXfield(40.0,units='K'))
    
    or
    
    >>> entry.sample.temperature = 40.0
    >>> entry.sample.temperature.units='K'
     
    """
    # Plotter to use for plot calls
    _plotter = PylabPlotter()

    def __init__(self, *items, **opts):
        if "name" in opts.keys():
            self.name = opts["name"]
            del opts["name"]
        self.entries = {}
        if "entries" in opts.keys():
            for k,v in opts["entries"].items():
                setattr(self, k, v)
            del opts["entries"]
        self.attrs = {}
        if "attrs" in opts.keys():
            self._setattrs(opts["attrs"])
            del opts["attrs"]
        for k,v in opts.items():
            setattr(self, k, v)
        if self.nxclass.startswith("NX"):
            if self.name == "unknown": self.name = self.nxclass[2:]
            try: # If one exists, set the class to a valid NXgroup subclass
                self.__class__ = globals()[self.nxclass]
            except KeyError:
                pass             
        for item in items:
            try:
                setattr(self, item.name, item)
            except AttributeError:
                raise NeXusError, "Non-keyword arguments must be valid NXobjects"

#    def __cmp__(self, other):
#        """Sort groups by their distances or names."""
#        try:
#            return cmp(self.distance, other.distance)
#        except KeyError:
#            return cmp(self.name, other.name)

    def __repr__(self):
        return "NXgroup('%s','%s')"%(self.nxclass,self.name)
        
    def _str_value(self,indent=0):
        return ""

    def __getattr__(self, key):
        """
        Provide direct access to groups via nxclass name.
        """
        if key.startswith('NX'):
            return self.component(key)
        elif key in self.entries:
            return self.entries[key]
        elif key in self.attrs:
            return self.attrs[key].nxdata
        raise KeyError(key+" not in "+self.nxclass+":"+self.name)

    def __setattr__(self, name, value):
        """
        Set an attribute as an object or regular Python attribute.
        
        It is assumed that attributes starting with 'nx' or '_' are regular 
        Python attributes. All other attributes are converted to valid NXobjects,
        with class NXfield, NXgroup, or a sub-class of NXgroup, depending on the
        assigned value.
        
        The internal value of the attribute name, i.e., 'name', is set to the 
        attribute name used in the assignment.  The parent group of the 
        attribute, i.e., 'group', is set to the parent group of the attribute.
        
        If the assigned value is a numerical (scalar or array) or string object,
        it is converted to an object of class NXfield, whose attribute, 'nxdata', 
        is set to the assigned value.
        """
        if isinstance(value, NXattr):
            self.attrs[name] = value
        elif name.startswith('nx') or name.startswith('_') or name in _nxattrs:
            object.__setattr__(self, name, value)
        elif isinstance(value, NXobject):
            value.group = self
            value.name = name
            self.entries[name] = value
        else:
            self.entries[name] = NXfield(value=value, name=name, group=self)

    def __getitem__(self, index):
        """
        Returns a slice from the NXgroup nxsignal attribute (if it exists) as
        a new NXdata group.
        
        In most cases, the slice values are applied to the NXfield nxdata array
        and returned within an NXfield object with the same metadata. However,
        if the array is one-dimensional and the index start and stop values 
        are real, the nxdata array is returned with values between the limits
        set by those axis values.
        This is to allow axis arrays to be limited by their actual value. This
        real-space slicing should only be used on monotonically increasing (or
        decreasing) one-dimensional arrays.
        """
        if not self.nxsignal: 
            raise NeXusError, "No plottable signal"
        if not hasattr(self,"nxclass"):
            raise NeXusError, "Indexing not allowed for groups of unknown class"
        if isinstance(index, int):
            axes = self.nxaxes
            axes[0] = axes[0][index]
            result = NXdata(self.nxsignal[index], axes)
            if self.nxerrors: result.errors = self.errors[index]
        elif isinstance(index, slice):
            axes = self.nxaxes
            axes[0] = axes[0][index]
            if isinstance(index.start, float) or isinstance(index.stop, float):
                index = slice(self.nxaxes[0].index(index.start), 
                              self.nxaxes[0].index(index.stop,max=True)+1)
                result = NXdata(self.nxsignal[index], axes)
                if self.nxerrors: result.errors = self.errors[index]
            else:
                result = NXdata(self.nxsignal[index], axes)
                if self.nxerrors: result.errors = self.errors[index]
        else:
            i = 0
            slices = []
            axes = self.nxaxes
            for ind in index:
                axes[i] = axes[i][ind]
                if isinstance(ind, slice) and \
                   (isinstance(ind.start, float) or isinstance(ind.stop, float)):
                    slices.append(slice(self.nxaxes[i].index(ind.start),
                                        self.nxaxes[i].index(ind.stop)))
                else:
                    slices.append(ind)
                i = i + 1
            result = NXdata(self.nxsignal.__getitem__(tuple(slices)), axes)
            if self.nxerrors: result.errors = self.errors.__getitem__(tuple(slices))
        axes = []
        for axis in result.nxaxes:
            if len(axis) > 1: axes.append(axis)
        result.nxsignal.axes = ":".join([axis.name for axis in axes])
        result.title = self.nxtitle
        return result

    def _setNXfield(self, value, name):
        if isinstance(value, list) or isinstance(value, tuple):
            setattr(self, name, NXfield(value=np.array(value), name=name))
            return name
        elif isinstance(value, np.ndarray):
            setattr(self, name, NXfield(value=value, name=name))
            return name
        elif isinstance(value, NXfield):
            if value.name == 'unknown': value.name = name
            setattr(self, value.name, value)
            return value.name
        else:
            raise NeXusError, "%s: '%s' should be a Numpy array or an NXfield" % (type(value), name)

    def insert(self, value, name='unknown'):
        """
        Adds an attribute to the group.
        
        If it is not a valid NeXus object (NXfield or NXgroup), the attribute 
        is converted to an NXfield.        
        """
        if isinstance(value, NXobject):
            if name == 'unknown': name = value.name
            if name in self.entries:
                raise NeXusError, "'%s' already exists in group" % name
            value.group = self
            self.entries[name] = value
        else:
            self.entries[name] = NXfield(value=value, name=name, group=self)

    def makelink(self, target):
        """
        Create a linked NXobject within the group.
        
        All attributes are inherited from the parent object including the name
        """
        if isinstance(target, NXobject):
            self.entries[target.name] = NXlink(target=target, name=target.name, group=self)
        else:
            raise NeXusError, "Link target must be an NXobject"

        
    def sum(self, axis=None):
        """
        Return the sum of the NXdata group using the Numpy sum method
        on the NXdata signal.

        The result contains a copy of all the metadata contained in
        the NXdata group. 
        """
        if not self.nxsignal: 
            raise NeXusError, "No signal to sum"
        if not hasattr(self,"nxclass"):
            raise NeXusError, "Summing not allowed for groups of unknown class"
        if axis is None:
            return self.nxsignal.sum()
        else:
            signal = NXfield(self.nxsignal.sum(axis), name=self.nxsignal.name)
            axes = self.nxaxes
            summedaxis = axes.pop(axis)
            units = ""
            if hasattr(summedaxis, "units"): units = summedaxis.units
            signal.long_name = "Integral from %s to %s %s" % \
                               (summedaxis[0], summedaxis[-1], units)
            average = NXfield(0.5*(summedaxis.nxdata[0]+summedaxis.nxdata[-1]), name=summedaxis.name)
            if units: average.units = units
            result = NXdata(signal, axes, average)
            if self.nxerrors:
                errors = np.sqrt((self.nxerrors.nxdata**2).sum(axis))
                result.errors = NXfield(errors, name="errors")
            result.title = self.nxtitle
            return result        

    def moment(self, order=1):
        """
        Return an NXfield containing the moments of the NXdata group 
        assuming the signal is one-dimensional.
        
        Currently, only the first moment has been defined. Eventually, the 
        order of the moment will be defined by the 'order' parameter.
        """
        if not self.nxsignal: 
            raise NeXusError, "No signal to calculate"
        elif len(self.nxsignal.shape) > 1:
            raise NeXusError, "Operation only possible on one-dimensional signals"
        elif order > 1:
            raise NeXusError, "Higher moments not yet implemented"
        if not hasattr(self,"nxclass"):
            raise NeXusError, "Operation not allowed for groups of unknown class"
        return (centers(self.nxsignal,self.nxaxes)*self.nxsignal).sum() \
                /self.nxsignal.sum()

    def component(self, nxclass):
        """
        Find all child objects that have a particular class.
        """
        return [E for _name,E in self.entries.items() if E.nxclass==nxclass]

    def signals(self):
        """
        Return a dictionary of NXfield's containing signal data.
        
        The key is the value of the signal attribute.
        """
        signals = {}
        for obj in self.entries.values():
            if 'signal' in obj.attrs:
                signals[obj.nxsignal.nxdata] = obj
        return signals                

    def _signal(self):
        """
        Return the NXfield containing the signal data.
        """
        for obj in self.entries.values():
            if 'signal' in obj.attrs and str(obj.signal) == '1':
                return self.entries[obj.name]
        return None

    def _axes(self):
        """
        Return a list of NXfields containing the axes.
        """
        try:
            return [getattr(self,name) for name in _getAxes(self.nxsignal.axes)]
        except KeyError:
            axes = {}
            for obj in self.entries:
                if 'axis' in getattr(self,obj).attrs:
                    axes[getattr(self,obj).axis] = getattr(self,obj)
            return [axes[key] for key in sorted(axes.keys())]

    def _errors(self):
        """
        Return the NXfield containing the signal errors.
        """
        try:
            return self.entries['errors']
        except KeyError:
            return None

    def _title(self):
        """
        Return the title as a string.
        
        If there is no title attribute in the string, the parent 
        NXentry group in the group's path is searched.
        """
        title = self.path
        if 'title' in self.entries:
            return str(self.title)
        else:
            obj = self
            while obj.group is not None:
                try:
                    if obj.nxclass == 'NXentry': return str(obj.title)
                except KeyError:
                    pass
                obj = obj.group
            return self.path        

    nxsignal = property(_signal, "Signal NXfield within group")
    nxaxes = property(_axes, "List of axes within group")
    nxerrors = property(_errors, "Errors NXfield within group")
    nxtitle = property(_title, "Title for group plot")
    
    def _fixaxes(self):
        """
        Remove length-one dimensions from plottable data
        """
        signal = self.nxsignal
        if isinstance(self.nxsignal, NXlinkfield): signal = signal.link
        shape = list(signal.shape)
        while 1 in shape: shape.remove(1)
        axes = []
        for axisname in _getAxes(signal.axes):
            axis = getattr(self,axisname)
            if len(axis) > 1: axes.append(axis)
        return signal.view().reshape(shape), axes

    def plot(self, signal=None, **opts):
        """
        Plot data contained within the group.
        
        Raises NeXusError if the data could not be plotted.
        """
        
        group = self
        if self.nxclass == "NXroot":
            group = group.NXentry[0]
        if group.nxclass == "NXentry":
            try:
                group = group.NXdata[0]
            except IndexError:
                raise NeXusError('No NXdata group found')
        
        # Find a plottable signal
        if signal is None:
            signal = group.nxsignal
            if not signal:
                raise NeXusError('No plottable signal defined')

        # Find errors
        errors= group.nxerrors
        
        # Find the associated axes
        axes = group.nxaxes
        
        # Provide a new view of the data if there is a dimension of length 1
        if 1 in signal.shape: signal, axes = group._fixaxes()

        # Construct title
        title = group.nxtitle
 
        # Plot with the available plotter
        group._plotter.plot(signal, axes, title, errors, **opts)


class NXlink(NXobject):
    """
    NeXus linked object.
    
    The real object will be accessible by following the link attribute.
    """
    nxclass = "NXlink"

    def __init__(self, target=None, name='link', group=None):
        self.group = group
        self.name = name
        if isinstance(target, NXobject):
            self._target = target.path
            if target.nxclass == "NXlink":
                raise NeXusError, "Cannot link to another NXlink object"
            elif target.nxclass == "NXfield":
                self.__class__ = NXlinkfield
            else:
                self.__class__ = NXlinkgroup
        else:
            self._target = target

    def __getattr__(self, key):
        try:
            try:
                return self.link.__dict__[key]
            except KeyError:
                return self.link.__getattr__(key)
        except KeyError:
            raise KeyError, (key+" not in %s" % self._target)

    def __repr__(self):
        return "NXlink('%s')"%(self._target)

    def __str__(self):
        return "NXlink('%s')"%(self._target)

    def _str_tree(self, indent=0, attrs=False, recursive=False):
        if self.link:
            return self.link._str_tree(indent, attrs, recursive)
        else:
            return " "*indent+self.name+' -> '+self._target

    def _link(self):
        link = self.head
        if link:
            try:
                for level in self._target[1:].split('/'):
                    link = link.entries[level]
                return link
            except AttributeError:
                return None
        else:
            return None

    def _attrs(self):
        return self.link.attrs

    def _entries(self):
        return self.link.entries

    link = property(_link, "Linked object")
    attrs = property(_attrs,doc="NeXus attributes for object")
    entries = property(_entries,doc="NeXus objects within group")    


class NXlinkfield(NXlink, NXfield):
    """
    NeXus linked NXfield.
    
    The real field will be accessible by following the link attribute.
    """
    def __setattr__(self, name, value):
        if name.startswith('_'):
            object.__setattr__(self, name, value)
        elif name == "name" or name == "nxclass" or name == "group":
            NXfield.__setattr__(self, name, value)
        elif self.link:
            self.link.__setattr__(name, value)

    def get(self, offset, size, units=""):
        """
        Get a slab from the data array.

        Offsets are 0-origin.  Shape can be inferred from the data.
        Offset and shape must each have one entry per dimension.
        
        If units are specified, convert the values to the given units
        before returning them.

        This operation should be performed in a "with group.data"
        conext.

        Raises ValueError cannot convert units.

        Corresponds to NXgetslab(handle,data,offset,shape)
        """
        if self.head._file:
            self.link.__enter__()
            value = self.link.head._file.getslab(offset,size)
            return self.link._converter(value,units)
            self.link.__exit__()
        else:
            raise NeXusError, "No input file specified for NXgetslab"
    
NXlinkdata = NXlinkfield # For backward compatibility

class NXlinkgroup(NXlink, NXgroup):
    """
    NeXus linked group.
    
    The real group will be accessible by following the link attribute.
    """
    def __setattr__(self, name, value):
        if name.startswith('_'):
            object.__setattr__(self, name, value)
        elif name == "name" or name == "nxclass" or name == "group":
            NXgroup.__setattr__(self, name, value)
        elif self.link:
            if not isinstance(value, NXobject): value = NXfield(value=value, name=name)
            value.group = self
            value.name = name
            object.__setattr__(self, name, value)
            setattr(self.link, name, value)


class NXentry(NXgroup):
    """
    NXentry group. This is a subclass of the NXgroup class.

    Each NXdata and NXmonitor object of the same name will be added
    together, raising an NeXusError if any of the groups do not exist 
    in both NXentry groups or if any of the NXdata additions fail. 
    The resulting NXentry group contains a copy of all the other metadata 
    contained in the first group. Note that other extensible data, such 
    as the run duration, are not currently added together. 

    See the NXgroup documentation for more details.
    """
    def __init__(self, *items, **opts):
        self.nxclass = "NXentry"
        NXgroup.__init__(self, *items, **opts)

    def __add__(self, other):
        """
        Add two NXentry objects       
        """
        result = NXentry(entries=self.entries, attrs=self.attrs)
        try:
            names = [group.name for group in self.component("NXdata")]
            for name in names:
                if isinstance(other.entries[name], NXdata):
                    result.entries[name] = self.entries[name] + other.entries[name]
                else:
                    raise KeyError
            names = [group.name for group in self.component("NXmonitor")]
            for name in names:
                if isinstance(other.entries[name], NXmonitor):
                    result.entries[name] = self.entries[name] + other.entries[name]
                else:
                    raise KeyError
            return result
        except KeyError:
            raise NeXusError, "Inconsistency between two NXentry groups"

    def __sub__(self, other):
        """
        Subtract two NXentry objects
        """
        result = NXentry(entries=self.entries, attrs=self.attrs)
        try:
            names = [group.name for group in self.component("NXdata")]
            for name in names:
                if isinstance(other.entries[name], NXdata):
                    result.entries[name] = self.entries[name] - other.entries[name]
                else:
                    raise KeyError
            names = [group.name for group in self.component("NXmonitor")]
            for name in names:
                if isinstance(other.entries[name], NXmonitor):
                    result.entries[name] = self.entries[name] - other.entries[name]
                else:
                    raise KeyError
            return result
        except KeyError:
            raise NeXusError, "Inconsistency between two NXentry groups"


class NXsubentry(NXentry):
    """
    NXsubentry group. This is a subclass of the NXsubentry class.

    See the NXgroup documentation for more details.
    """
    def __init__(self, *items, **opts):
        self.nxclass = "NXsubentry"
        NXgroup.__init__(self, *items, **opts)


class NXdata(NXgroup):
    """
    NXdata group. This is a subclass of the NXgroup class.
    
    The constructor assumes that the first argument contains the signal and
    the second contains either the axis, for one-dimensional data, or a list
    of axes, for multidimenional data. These arguments can either be NXfield 
    objects or Numpy arrays, which are converted to NXfield objects with default 
    names.
    
    Various arithmetic operations (addition, subtraction, multiplication, 
    and division) have been defined for combining NXdata groups with other
    NXdata groups, Numpy arrays, or constants, raising a NeXusError if the 
    shapes don't match. Data errors are propagated in quadrature if 
    they are defined, i.e., if the 'nexerrors' attribute is not None, 

    Attributes
    ----------
    nxsignal : The NXfield containing the attribute 'signal' with value 1
    nxaxes   : A list of NXfields containing the signal axes
    nxerrors : The NXfield containing the errors

    Methods
    -------
    plot(self, over=False, log=False, **opts)
        Plot the NXdata group using the defined signal and axes. Valid
        Matplotlib parameters, specifying markers, colors, etc, can be 
        specified using the 'opts' dictionary.    
    
    moment(self, order=1)
        Calculate moments of the NXdata group. This assumes that the 
        signal is one-dimenional. Currently, only the first moment is 
        implemented.

    Examples
    --------
    >>> x = np.linspace(0, 2*np.pi, 101)
    >>> line = NXdata(sin(x), x)
    data:NXdata
      signal = float64(101)
        @axes = x
        @signal = 1
      x = float64(101)
    >>> X, Y = np.meshgrid(x, x)
    >>> z = NXfield(sin(X) * sin(Y), name='intensity')
    >>> entry = NXentry()
    >>> entry.grid = NXdata(z, (x, x))
    >>> grid.tree()
    entry:NXentry
      grid:NXdata
        axis1 = float64(101)
        axis2 = float64(101)
        intensity = float64(101x101)
          @axes = axis1:axis2
          @signal = 1    

    See the NXgroup documentation for more details.
    """
    def __init__(self, signal=None, axes=(), *items, **opts):
        self.nxclass = "NXdata"
        NXgroup.__init__(self, *items, **opts)
        if signal is not None:
            signalname = self._setNXfield(signal, "signal")
            self.entries[signalname].signal = 1
            if axes is not None:
                if isinstance(axes,tuple) or isinstance(axes,list):
                    axisname = {}
                    i = 0
                    for axis in axes:
                        i = i + 1
                        axisname[i] = self._setNXfield(axis, "axis%s" % i)
                    self.entries[signalname].axes = ":".join(axisname.values())
                else:
                    axisname = self._setNXfield(axes, 'x')
                    self.entries[signalname].axes = axisname

    def __add__(self, other):
        """
        Define a method for adding a NXdata group to another NXdata group
        or to a number. Only the signal data is affected.

        The result contains a copy of all the metadata contained in
        the first NXdata group. The module checks that the dimensions are 
        compatible, but does not check that the NXfield names or values are
        identical. This is so that spelling variations or rounding errors
        do not make the operation fail. However, it is up to the user to 
        ensure that the results make sense.
        """
        result = NXdata(entries=self.entries, attrs=self.attrs)
        if isinstance(other, NXdata):
            if self.nxsignal and self.nxsignal.shape == other.nxsignal.shape:
                result.entries[self.nxsignal.name] = self.nxsignal + other.nxsignal
                if self.nxerrors:
                    if other.nxerrors:
                        result.errors = np.sqrt(self.errors.nxdata**2+other.errors.nxdata**2)
                    else:
                        result.errors = self.errors
                return result
        elif isinstance(other, NXgroup):
            raise NeXusError, "Cannot add two arbitrary groups"
        else:
            result.entries[self.nxsignal.name] = self.nxsignal + other
            result.entries[self.nxsignal.name].name = self.nxsignal.name
            return result
        
    def __sub__(self, other):
        """
        Define a method for subtracting a NXdata group or a number from 
        the NXdata group. Only the signal data is affected.

        The result contains a copy of all the metadata contained in
        the first NXdata group. The module checks that the dimensions are 
        compatible, but does not check that the NXfield names or values are
        identical. This is so that spelling variations or rounding errors
        do not make the operation fail. However, it is up to the user to 
        ensure that the results make sense.
        """
        result = NXdata(entries=self.entries, attrs=self.attrs)
        if isinstance(other, NXdata):
            if self.nxsignal and self.nxsignal.shape == other.nxsignal.shape:
                result.entries[self.nxsignal.name] = self.nxsignal - other.nxsignal
                if self.nxerrors:
                    if other.nxerrors:
                        result.errors = np.sqrt(self.errors.nxdata**2+other.errors.nxdata**2)
                    else:
                        result.errors = self.errors
                return result
        elif isinstance(other, NXgroup):
            raise NeXusError, "Cannot subtract two arbitrary groups"
        else:
            result.entries[self.nxsignal.name] = self.nxsignal - other
            result.entries[self.nxsignal.name].name = self.nxsignal.name
            return result
        
    def __mul__(self, other):
        """
        Define a method for multiplying the NXdata group with a NXdata group
        or a number. Only the signal data is affected.

        The result contains a copy of all the metadata contained in
        the first NXdata group. The module checks that the dimensions are 
        compatible, but does not check that the NXfield names or values are
        identical. This is so that spelling variations or rounding errors
        do not make the operation fail. However, it is up to the user to 
        ensure that the results make sense.
        """
        result = NXdata(entries=self.entries, attrs=self.attrs)
        if isinstance(other, NXdata):

            # error here signal not defined in this scope
            #if self.nxsignal and signal.shape == other.nxsignal.shape:
            if self.nxsignal and self.nxsignal.shape == other.nxsignal.shape:
                result.entries[self.nxsignal.name] = self.nxsignal * other.nxsignal
                if self.nxerrors:
                    if other.nxerrors:
                        result.errors = np.sqrt((self.errors.nxdata*other.nxsignal.nxdata)**2+
                                                (other.errors.nxdata*self.nxsignal.nxdata)**2)
                    else:
                        result.errors = self.errors
                return result
        elif isinstance(other, NXgroup):
            raise NeXusError, "Cannot multiply two arbitrary groups"
        else:
            result.entries[self.nxsignal.name] = self.nxsignal * other
            result.entries[self.nxsignal.name].name = self.nxsignal.name
            if self.nxerrors:
                result.errors = self.errors * other
            return result

    def __rmul__(self, other):
        """
        Define a method for multiplying NXdata groups.

        This variant makes __mul__ commutative.
        """
        return self.__mul__(other)
        
    def __div__(self, other):
        """
        Define a method for dividing the NXdata group by a NXdata group
        or a number. Only the signal data is affected.

        The result contains a copy of all the metadata contained in
        the first NXdata group. The module checks that the dimensions are 
        compatible, but does not check that the NXfield names or values are
        identical. This is so that spelling variations or rounding errors
        do not make the operation fail. However, it is up to the user to 
        ensure that the results make sense.
        """
        result = NXdata(entries=self.entries, attrs=self.attrs)
        if isinstance(other, NXdata):
            if self.nxsignal and self.nxsignal.shape == other.nxsignal.shape:
                # error here, signal and othersignal not defined here
                #result.entries[self.nxsignal.name] = signal / othersignal
                result.entries[self.nxsignal.name] = self.nxsignal / other.nxsignal
                resultvalues = result.entries[self.nxsignal.name].nxdata
                if self.nxerrors:
                    if other.nxerrors:
                        result.errors = (np.sqrt(self.errors.nxdata**2 + 
                                         (resultvalues*other.errors.nxdata)**2)
                                         / other.nxsignal)                                              
                    else:
                        result.errors = self.errors
                return result
        elif isinstance(other, NXgroup):
            raise NeXusError, "Cannot divide two arbitrary groups"
        else:
            result.entries[self.nxsignal.name] = self.nxsignal / other
            result.entries[self.nxsignal.name].name = self.nxsignal.name
            if self.nxerrors: result.errors = self.errors / other
            return result


class NXmonitor(NXdata):
    """
    NXmonitor group. This is a subclass of the NXdata class.
    
    See the NXdata and NXgroup documentation for more details.
    """
    def __init__(self, signal=None, axes=(), *items, **opts):
        NXdata.__init__(self, signal=signal, axes=axes, *items, **opts)
        self.nxclass = "NXmonitor"
        if "name" not in opts.keys():
            self.name = "monitor"


class NXlog(NXgroup):
    """
    NXlog group. This is a subclass of the NXgroup class.
    
    Methods
    -------
    plot(self, **opts)
        Plot the logged values against the elapsed time. Valid
        Matplotlib parameters, specifying markers, colors, etc, can be 
        specified using the 'opts' dictionary.
    
    See the NXgroup documentation for more details.
    """
    def __init__(self, *items, **opts):
        self.nxclass = "NXlog"
        NXgroup.__init__(self, *items, **opts)
    
    def plot(self, **opts):
        axis = [self.time]
        title = "%s Log" % self.value.name.upper()
        self._plotter.plot(self.value, axis, title, **opts)


#Add remaining base classes as subclasses of NXgroup and append to __all__
for _class in _nxclasses:
    if _class not in globals():
        globals()[_class]=type(_class,(NXgroup,),dict(nxclass=_class))
    __all__.append(_class)


def centers(signal, axes):
    """
    Return the centers of the axes regardless if the axes contain
    bin boundaries or centers.
    """
    def findc(axis, dimlen):
        if axis.shape[0] == dimlen+1:
            return (axis.nxdata[:-1] + axis.nxdata[1:])/2
        else:
            assert axis.shape[0] == dimlen
            return axis.nxdata
    return [findc(a,signal.shape[i]) for i,a in enumerate(axes)]

def setmemory(value):
    """
    Set the memory limit for data arrays (in MB).
    """
    global NX_MEMORY
    NX_MEMORY = value

def label(field):
    """
    Construct a label for a data field suitable for use on a graph axis.
    """
    if hasattr(field,'long_name'):
        return field.long_name
    elif hasattr(field,'units'):
        return "%s (%s)"%(field.name,field.units)
    else:
        return field.name

def imshow_irregular(x,y,z):
    import pylab
    from matplotlib.ticker import FormatStrFormatter
    ax = pylab.gca()
    im = pylab.mpl.image.NonUniformImage(ax, extent=(x[0],x[-1],y[0],y[-1]), origin=None)
    im.set_data(x,y,z)
    ax.images.append(im)
    ax.set_xlim(x[0],x[-1])
    ax.set_ylim(y[0],y[-1])
    pylab.colorbar(im) #format=FormatStrFormatter('$10^{%d}$')
    pylab.gcf().canvas.draw_idle()

# File level operations
def load(filename, mode='r'):
    """
    Read a NeXus file, returning a tree of objects
    
    This is aliased to 'read' because of potential name clashes with Numpy
    """
    file = NeXusTree(filename,mode)
    tree = file.readfile()
    file.close()
    return tree

#Definition for when there are name clashes with Numpy
read = load
__all__.append('read')

def save(filename, group, format='w5'):
    """
    Write a NeXus file from a tree of objects
    """
    if group.nxclass == "NXroot":
        tree = group
    elif group.nxclass == "NXentry":
        tree = NXroot(group)
    else:
        tree = NXroot(NXentry(group))
    file = NeXusTree(filename, format)
    file.writefile(tree)
    file.close()

def tree(file):
    """
    Read and summarize the named nexus file.
    """
    nxfile = load(file)
    nxfile.tree()

def demo(argv):
    """
    Process command line commands in argv.  argv should contain
    program name, command, arguments, where command is one of
    the following:
        copy fromfile.nxs tofile.nxs
        ls f1.nxs f2.nxs ...
    """
    if len(argv) > 1:
        op = argv[1]
    else:
        op = 'help'
    if op == 'ls':
        for f in argv[2:]: dir(f)
    elif op == 'copy' and len(argv)==4:
        tree = load(argv[2])
        save(argv[3], tree)
    elif op == 'plot' and len(argv)==4:
        tree = load(argv[2])
        for entry in argv[3].split('.'):
            tree = getattr(tree,entry)
        tree.plot()
        tree._plotter.show()
            
    else:
        usage = """
usage: %s cmd [args]
    copy fromfile.nxs tofile.nxs
    ls *.nxs
    plot file.nxs entry.data
        """%(argv[0],)
        print usage
        

if __name__ == "__main__":
    import sys
    demo(sys.argv)

